#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import inspect
import numpy as np
import sys
from pathlib import PurePath
#Start to import others .py code from others directory
lcode = PurePath(__file__)
datalocation = lcode.parents[2].as_posix() + '/'
export_location = lcode.parents[2].as_posix() + '/Data_Raw/'
if  (lcode.parents[1].as_posix() in sys.path)  is False:
    sys.path.insert(0,  lcode.parents[1].as_posix())



def retrieve_name(var):
    """
    Retrieve the name of the variable by searching from the outermost frame inward..
    """
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    return [var_name for var_name, var_val in callers_local_vars if var_val is var]

class NpEncoder(json.JSONEncoder):
    """
    Class for encoding a NumPy array to a JSON file.
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)
    
def save_json_file(filename, *args):
    """
    
    Save a JSON file with the specified filename and the given arguments. 
    'args' is a list of lists, each containing two elements.
    
    """

    json_data = {}
    if len(args) > 0:
        for arg in args:
            if isinstance(arg, list) == False:
                print('major issue for save please check: ' + str(arg))
            else:   
                json_data[arg[0]] = arg[1]
            
        with open(filename,'w') as f:
            try:
                json.dump(json_data, f, cls=NpEncoder)
            except TypeError:
            
                print('ERROR json: for' + str(json_data))
                f.close()
            f.close()
    else:
        print("nothing to save")

 
def generate_filename(identity,methodname,Nbsteps,thetas, K, suffix='.json',nrun=0):
    
    """
    Generate a filename for the given identity, methodname, Nbsteps, thetas, K, suffix, and nrun.
    This function is used to create an appropriate filename for a JSON file to save data in the 'Data' folder.

    Parameters:
    - 'identity': Additional information that can be manually added to the filename.
    - 'methodname': The name of the algorithm used.
    - 'Nbsteps': The number of steps in the simulation.
    - 'thetas': A list of the means of the arms in the simulation. If 'thetas[0]' is -1, thetas are assumed to be uniformly distributed.
    - 'K': The number of arms in the simulation.
    - 'suffix': The suffix of the file (default is '.json').
    - 'nrun': The number of runs for the simulation (default is 0).
   
     """
    filename = datalocation + 'Data/' + identity +'method_'+ str(methodname) +'_time_' + str(Nbsteps)  
    if nrun != -1:
        filename += '_nrun_' + str(nrun)
    
    if thetas[0] == -1:
        filename += '_theta-1_'
    elif thetas[0] == -2:
        filename += '_theta-2_'
    else:
        filename += '_thetalist_'
    
    filename += '_K_' + str(K)

    filename += suffix
    return filename

def generate_server_filename(numprocess,fileidentity,methodname,Nbsteps,thetas, K, suffix='.json',nrun=0):
    """
    Generate a filename for server utility. This function is used to create a suitable filename for a JSON file in order to save data in the 'Data' folder when the simulation is run on a server.

    Parameters:
    - 'numprocess': The number of processes (as a job or task in SLURM).
    - 'fileidentity': Additional information to be included in the filename.
    - 'methodname': The name of the algorithm used.
    - 'Nbsteps': The number of steps in the simulation.
    - 'thetas': A list of the means of the arms in the simulation. If 'thetas[0]' is -1, thetas are assumed to be uniformly distributed.
    - 'K': The number of arms in the simulation.
    - 'suffix': The suffix of the file (default is '.json').
    - 'nrun': The number of runs for the simulation (default is 0).
    
    """

    filename = export_location + fileidentity +'method_'+ str(methodname) +'_time_' + str(Nbsteps)  
    if nrun != -1:
        filename += '_nrun_' + str(nrun)
    
    if thetas[0] == -1:
        filename += '_theta-1_'
    elif thetas[0] == -2:
        filename += '_theta-2_'
    else:
        filename += '_thetalist_'
    
    filename += '_K_' + str(K)

    filename += '_id_' + str(numprocess)
    filename += suffix
    
    return filename



